import { ListenerRegistry } from "./util/ListenerRegistry";
/**
 * A base interface for entities, which are just normal JavaScript objects with
 * any number of properties, each of which represents a single component. When
 * using hmecs, it is recommended to create your own type representing your
 * game's entities, and pass that to `createWorld` for full typing support.
 */
export interface IEntity {
    id?: number;
}
/**
 * For situations where no entity type argument is passed to createWorld, we'll
 * default to an untyped entity type that can hold any component.
 */
export declare type UntypedEntity = {
    [components: string]: ComponentData;
} & IEntity;
/**
 * Component names are just strings/object property names.
 */
export declare type ComponentName<T extends IEntity> = keyof T;
/**
 * The data of a component can be literally anything. Good times!
 */
export declare type ComponentData = any;
/**
 * miniplex lets you define archetypes of entities. For each archetype,
 * it will automatically create an index; entities matching this archetype
 * will automatically be added to the index, entities no longer matching
 * the archetype will automatically be removed.
 */
export declare type Archetype<T extends IEntity> = {
    all?: ComponentName<T>[];
    any?: ComponentName<T>[];
    none?: ComponentName<T>[];
};
/**
 * An archetype index represents a mapping between archetypes to simple lists
 * of entities (of this archetype.) It uses the archetypes' object identities
 * as keys.
 */
export declare type ArchetypeIndex<T extends IEntity> = Map<Archetype<T>, T[]>;
declare type FunctionWithQueuedVariant<ImmediateFunction extends {
    (...args: any[]): any;
}, QueuedReturnType = void> = ImmediateFunction & {
    queued: (...args: Parameters<ImmediateFunction>) => QueuedReturnType;
};
declare type Listeners<T> = {
    listeners: {
        archetypeChanged: Map<Archetype<T>, ListenerRegistry>;
    };
};
declare type ArchetypeOrComponentList<T> = ComponentName<T>[] | [Partial<Archetype<T>>];
export declare type World<T extends IEntity> = {
    entities: T[];
    addEntity: FunctionWithQueuedVariant<(entity: T) => T, T>;
    removeEntity: FunctionWithQueuedVariant<(entity: T) => void>;
    addComponent: FunctionWithQueuedVariant<(<U extends keyof T>(entity: T, name: U, data: T[U]) => void)>;
    removeComponent: FunctionWithQueuedVariant<(entity: T, ...names: ComponentName<T>[]) => void>;
    createArchetype: (...input: ArchetypeOrComponentList<T>) => Archetype<T>;
    get: (...input: ArchetypeOrComponentList<T>) => T[];
    getOne: (...input: ArchetypeOrComponentList<T>) => T | undefined;
    flushQueue: () => void;
    clear: () => void;
} & Listeners<T>;
/**
 * Create an ECS world.
 */
export declare function createWorld<T extends IEntity = UntypedEntity>(): World<T>;
export {};
