import { commandQueue } from "./util/commandQueue";
import { entityIsArchetype } from "./util/entityIsArchetype";
import { idGenerator } from "./util/idGenerator";
import { ListenerRegistry } from "./util/ListenerRegistry";
import { memoizedMap } from "./util/memoizedMap";
/**
 * Create an ECS world.
 */
export function createWorld() {
    /**
     * An array holding all entities known to this world.
     */
    const entities = new Array();
    let nextId = idGenerator(1);
    /**
     * A register of archetype-specific indices.
     */
    const archetypes = new Map();
    const listeners = {
        archetypeChanged: new Map()
    };
    /**
     * A useful queue of commands to run.
     */
    const queuedCommands = commandQueue();
    /**
     * A memoization cache for archetypes to help with maintaining
     * referential integrity.
     */
    const memoizedArchetypes = memoizedMap();
    const normalizeComponentList = (names) => names?.filter((n) => typeof n === "string" && n !== "").sort();
    /**
     * Normalize an archetype by sorting the component names it references.
     */
    const normalizeArchetype = (archetype) => ({
        all: archetype.all ? normalizeComponentList(archetype.all) : undefined,
        any: archetype.any ? normalizeComponentList(archetype.any) : undefined,
        none: archetype.none ? normalizeComponentList(archetype.none) : undefined
    });
    const memoizeArchetype = (archetype) => memoizedArchetypes.fetch(archetype, () => archetype);
    function createArchetype(...input) {
        /* Normalize and memoize archetype */
        const archetype = memoizeArchetype(normalizeArchetype(typeof input[0] === "string"
            ? { all: input }
            : input[0]));
        /* Create an index if we need to. */
        if (!archetypes.has(archetype)) {
            /* create an index, and index existing entities */
            archetypes.set(archetype, entities.filter((entity) => entityIsArchetype(entity, archetype)));
            /* Create listeners */
            listeners.archetypeChanged.set(archetype, new ListenerRegistry());
        }
        return archetype;
    }
    function indexEntity(entity) {
        /* We're going to loop through all known archetypes, check if the entity matches,
           and then remove/add it accordingly. */
        for (const [archetype, index] of archetypes.entries()) {
            const isArchetype = entityIsArchetype(entity, archetype);
            const pos = index.indexOf(entity, 0);
            if (isArchetype && pos < 0) {
                index.push(entity);
                listeners.archetypeChanged.get(archetype).invoke();
            }
            else if (!isArchetype && pos >= 0) {
                index.splice(pos, 1);
                listeners.archetypeChanged.get(archetype).invoke();
            }
        }
    }
    function removeEntityFromAllIndices(entity) {
        for (const [archetype, index] of archetypes.entries()) {
            const pos = index.indexOf(entity, 0);
            if (pos >= 0) {
                index.splice(pos, 1);
                listeners.archetypeChanged.get(archetype).invoke();
            }
        }
    }
    function get(...input) {
        return archetypes.get(createArchetype(...input));
    }
    function getOne(...input) {
        return get(...input)[0];
    }
    const addEntity = (entity) => {
        /* If there already is an ID, raise an error. */
        if ("id" in entity)
            throw "Attempted to add an entity that aleady had an 'id' component.";
        /* Assign an ID */
        entity.id = nextId();
        /* Store the entity... */
        entities.push(entity);
        /* ...and add it to relevant indices. */
        indexEntity(entity);
        return entity;
    };
    const removeEntity = (entity) => {
        /* Remove entity from all indices */
        removeEntityFromAllIndices(entity);
        /* Remove its id component */
        delete entity.id;
        /* Remove it from our global list of entities */
        const pos = entities.indexOf(entity, 0);
        entities.splice(pos, 1);
    };
    const addComponent = (entity, name, data) => {
        entity[name] = data;
        indexEntity(entity);
    };
    const removeComponent = (entity, ...components) => {
        components.forEach((name) => delete entity[name]);
        indexEntity(entity);
    };
    /* Queued versions of mutation functions */
    addEntity.queued = (entity) => {
        queuedCommands.add(() => addEntity(entity));
        return entity;
    };
    removeEntity.queued = (entity) => {
        queuedCommands.add(() => removeEntity(entity));
    };
    addComponent.queued = (entity, name, data) => {
        queuedCommands.add(() => addComponent(entity, name, data));
    };
    removeComponent.queued = (entity, ...names) => {
        queuedCommands.add(() => removeComponent(entity, ...names));
    };
    /**
     * Clear the entire world by discarding all entities and indices.
     */
    function clear() {
        /* Remove all entities */
        entities.length = 0;
        /* Remove all archetype indices */
        archetypes.clear();
        /* Remove all listeners */
        listeners.archetypeChanged.clear();
        /* Clear queue */
        queuedCommands.clear();
    }
    return {
        entities,
        listeners,
        createArchetype,
        get,
        getOne,
        addEntity,
        removeEntity,
        addComponent,
        removeComponent,
        flushQueue: queuedCommands.flush,
        clear
    };
}
